#!/usr/bin/perl -w

#Copyright (C) 2006-2008 Keio University
#(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)
#
#This file is part of Murasaki.
#
#Murasaki is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#Murasaki is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with Murasaki.  If not, see <http://www.gnu.org/licenses/>.

use strict;

use Getopt::Long;
use Pod::Usage;
use File::Basename;
#use Data::Dump qw {dump};

use strict;

BEGIN {
  our $root=(fileparse($0))[1];
  $root=$ENV{MURASAKI} if $ENV{MURASAKI};
  unshift(@INC,$root);
  require "$root/config.pl";
  require "$root/common.pl";
  unshift(@INC,$root);
}
use KOG;

my ($man,$help,$makecds);
my (@seqnames,$output_opt);
my $opterr=
  GetOptions('help|?' => \$help, man => \$man,
	    'output=s'=>\$output_opt);
pod2usage(1) if $help;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;

pod2usage({-verbose=>1,-exitval=>2,-message=>'Need some input file...'}) if $#ARGV<0;

my $round;
foreach my $inf (@ARGV){
  my ($iprefix,$dir,$ext)=fileparse($inf,qr{\.anchors});
  die "Need a .anchors file." unless $ext eq ".anchors";
  my $seqfile="$dir/$iprefix.seqs";
  die "Need $seqfile" unless -f $seqfile;
  open(my $infh,$inf) or die "Couldn't open $inf";
  my @seqs=map { {file=>$_} } (split(/\n/,slurp($seqfile)));
  foreach my $seq (@seqs){
    print "Loading regions for $seq->{file}\n";
    $seq->{regions}=[getRegions($seq->{file})];
  }

  my $outf=($output_opt and $round) ? "$output_opt.$round":$output_opt;
  $outf="$inf.osfinder-in" unless $outf;
  print STDERR "Warning: Overwriting $outf" if -f $outf;
  open(my $ofh,">$outf");
  while(my $line=<$infh>){
    my @anchors=getAnchors($line);
    die "Illegal number of anchors (".scalar(@anchors).")" unless scalar(@anchors)==scalar(@seqs);
    print $ofh join("\t",
		    map {fixInRegion($anchors[$_],$seqs[$_]->{regions})} 0..$#anchors)."\n";
  }
}

sub fixInRegion {
  my ($anchor,$regions)=@_;
  return (1,$anchor->{start},$anchor->{stop}) unless $regions;
  die "programming error ftl!" unless ref($regions) eq 'ARRAY';
  foreach my $i (0..$#{$regions}){
    my $region=$$regions[$i];
    if($anchor->{start}>=$region->{start} and $anchor->{stop}<=$region->{stop}){
      return ($i+1,$anchor->{start}-$region->{start},$anchor->{stop}-$region->{start},$anchor->{sign});
    }
  }
  die "Uh oh. No matching region found for anchor: ".anchorToString($anchor);
}

sub anchorToString {
  my ($anchor)=@_;
  return join("\t",$anchor->{start},$anchor->{stop},$anchor->{sign});
}

sub getAnchors {
  my ($line)=@_;
  my @d=split(/\t/,$line);
  my @anchors;
  while(@d){
    my ($start,$stop,$sign)=map {shift(@d)} (1..3);
    ($start,$stop)=(-$start,-$stop) if($start<0);
    ($start,$stop)=($stop,$start) if $stop<$start; #be flexible
    push(@anchors,{start=>$start,stop=>$stop,sign=>$sign});
  }
  return @anchors
}

sub getRegions {
  my ($file)=@_;
  die "No file specified?" unless $file;
  return undef unless $file=~m/\.stitch$/;
  open(my $fh,$file) or die "Couldn't open stitch sub-file $file";
  my @res;
  while(<$fh>){
    my @dat=split(/\t/);
    push(@res,{file=>$dat[0],length=>$dat[1],start=>$dat[2],stop=>$dat[3]});
  }
  return @res;
}

sub slurp {
  local $/;
  open(my $fh,"@_") or return;
  return <$fh>;
}
