/*
Copyright (C) 2006-2008 Keio University
(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)

This file is part of Murasaki.

Murasaki is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Murasaki is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Murasaki.  If not, see <http://www.gnu.org/licenses/>.
*/

/////////////////////
// murasaki config options manager
////////////////////

#include "options.h"
#include "murasaki.h"
#include "murasaki_mpi.h"
#include "timing.h"
#include "dinkymath.h"
#include "sequence.h"

#include <boost/filesystem/operations.hpp> //graceful handling of directories
#include <boost/lexical_cast.hpp>
#include <functional>
#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <string.h>
#include <getopt.h>


#ifdef MURASAKI_MPI
#include "mpi.h"
#endif

using namespace std;
using namespace boost::filesystem;

Options::Options():
  verbose(0),quick_hash(3),hitfilter(0),histogram(1),
  user_seed(false),rand_seed(time(NULL)),repeatMask(false),
  skip1to1(false),skipFwd(false),skipRev(false),doubleHash(false),
  hashOnly(false),hashCache(false),bitscore(true),
  hashPref(ArrayHash_t),alignment_name("test"),output_dir("output"),
  seedfilter(0),hashfilter(0),mergeFilter(0),
  joinNeighbors(false),joinSpec(-1),joinDist(-1),
  hashSkip(1),
  auto_hashbits(true),
  dumpRegions(false),dumpEntropy(false),leaveRecords(false),
  targetMemory(sysinfo.totalMemory*90/100),userSetMemory(false),
  retainMembers(true),commutative(true),useHumanTime(true),
  repeatMap(true),anchorProgressCheck(0),
  extractInOrder(false),tfidf(true),reverseOtf(true),
  rifts(0),
  //mpi related
  mpi_hashers(0),mpi_noCake(false),mpi_fileDistro(true),
  mpi_anchorInSpareTime(true),mpi_maxBuffers(0),mpi_bigFirst(false),
  mpi_hostbalance(true),mpi_memoryBalance(true),mpi_distMerge(true),mpi_distCollect(true),
  use_shm_mmap(false),mmap_writePerHost(false),
#if defined(MURASAKI_MPI) && defined(USE_SHM_SYSV)
  use_shm_sysv(true)
#else
  use_shm_sysv(false)
#endif
{
  try {
    if(exists("~/.murasaki")){
	config_file="~/.murasaki";
	loadConfig(config_file); //load both!
    }
  }catch(exception &e){}
  try {
    if(exists("murasaki.cfg")){
      config_file="murasaki.cfg";
      loadConfig(config_file);//load both!
    }
  }catch(exception &e){}
}

void Options::preInit(){
#ifdef MURASAKI_MPI
  if(mpi_capable)
    mpi_maxBuffers=1024/mpi_procs+1;
#endif
}

void Options::commandline_opts(int argc,char **argv){
  using namespace boost;
  int optc;
  while(1){
    //options struct:
    // name, has_arg, store_pointer, return_value
    static struct option long_options[] = {
      {"verbose",0,0,'v'},
      {"directory",1,0,'d'},
      {"pattern",1,0,'p'},
      {"name",1,0,'n'},
      {"help",0,0,'?'},
      {"version",0,0,'V'},
      {"quickhash",1,0,'q'},
      {"hashbits",1,0,'b'},
      {"hitfilter",1,0,'h'},
      {"histogram",2,0,'H'},
      {"hashtype",1,0,'t'},
      {"repeatmask",2,0,'r'},
      {"rseed",1,0,'s'},
      {"skip1to1",2,0,'1'},
      {"skipfwd",2,0,'F'},
      {"skiprev",2,0,'R'},
      {"doublehash",2,0,'D'},
      {"hashonly",2,0,'Q'},
      {"hashcache",2,0,'c'},
      {"seedfilter",1,0,'f'},
      {"hashfilter",1,0,'m'},
      {"bitscore",2,0,'B'},
      {"join",2,0,'j'},
      {"hashskip",2,0,'S'},
      {"dumpregions",2,0,'G'},
      {"dumpentropy",2,0,'E'},
      {"memory",1,0,'M'},
      {"seedterms",2,0,'T'},
      {"commutative",2,0,'C'},
      {"hashers",1,0,'A'},
      {"nocake",0,0,'K'},
      {"mpidistro",0,0,'L'},
      {"sectime",0,0,'e'},
      {"waittoanchor",2,0,'w'},
      {"mergefilter",1,0,'Y'},
      {"mmap",2,0,'P'},
      {"mmapperhost",2,0,'W'},
      {"sysvipc",2,0,'y'},
      {"nobuffer",0,0,'U'},
      {"buffers",1,0,'u'},
      {"bigfirst",0,0,'I'},
      {"repeatmap",2,0,'i'},
      {"hostbalance",2,0,'l'},
      {"extractinorder",2,0,'x'},
      {"tfidf",2,0,'k'},
      {"reverseotf",2,0,'o'},
      {"anchorprogress",1,0,'g'},
      {"memorybalance",2,0,'a'},
      {"leaverecords",2,0,'J'}, //J for junk
      {"distmerge",2,0,'<'}, //running out of letters!!!!
      {"distcollect",2,0,'>'}, 
      {"rifts",2,0,'%'},
      {0,0,0,0}
    };
    int longindex=0;
    string prefreq;
    optc=getopt_long(argc,argv,"G::J::W::MF::R::7D::Q::B::T::KL::w::P::y::UIE::Y:A:C::M:S:j:m:f:c::d:s:H::b:o::p:n:q:h:V?t:r::1::eu:i::l::vx::k::g:a::%:",long_options,&longindex);
    if(optc==-1)break;
    switch(optc){
    case 'j':
      if(optarg && (strcmp(optarg,"0"))){
	string optstr(optarg);
	if(optstr==string("0") || isNo(optarg)){
	  joinNeighbors=false;
	  break;
	}
      }
      joinNeighbors=true;
      joinSpec=-1;
      if(optarg && !sscanf(optarg,"%f",&joinSpec)){
	cerr << "Could not parse join specification.\n";
	cerr << program_help();exit(-1);
      }
      break;
    case 'S':
      if(optarg){
	if(!sscanf(optarg,"%d",&hashSkip)){
	  cerr << "Could not parse hashskip specification.\n";
	  cerr << program_help();exit(-1);
	}
      }else{
	hashSkip++;
      }
      if(hashSkip<=0){
	cerr << "Skip has to be at least 1.\n";
	exit(-1);
      }
      break;
    case 'A':
#ifdef MURASAKI_MPI      
      if(mpi_capable){
	if(optarg){
	  double percent;
	  if(sscanf(optarg,"%lf",&percent))
	    if(percent>=0 && percent<1){
	      mpi_hashers=(int)(percent*(double)mpi_procs);
	      if(mpi_hashers==0)
		mpi_hashers=1;
	    }else{
	      if(!sscanf(optarg,"%d",&mpi_hashers)){
		cerr << "Could not parse hashers specification.\n";
		cerr << program_help();exit(-1);
	      }
	    }
	}
	if(hashSkip<=0){
	  cerr << "Hashers has to be at least 1.\n";
	  exit(-1);
	}
      }else{
#endif
	cerr << "Option --hashers has no effect without MPI"<<endl;
#ifdef MURASAKI_MPI
      }
#endif
      break;

    case 'c':
      hashCache=true;
      if(optarg) cache_dir=string(optarg)+"/";
      else cache_dir="cache/";
      break;
    case 'v': verbose++;break;
    case 'V': cout << program_version();exit(-1);break;
    case 'd': output_dir=string(optarg);break;
    case 'p': pattern_src=string(optarg);break;
    case 'n': alignment_name=string(optarg);break;
    case 'J': parseYesNo(optarg,leaveRecords,"leaverecords");break;
    case 'q':
      if(!sscanf(optarg,"%d",&quick_hash)){
	cerr << "Could not parse quickhash specification.\n";
	cerr << program_help();exit(-1);
      }
      if(quick_hash>3 || quick_hash<0){
	cerr << "Invalid quickhash specification ("<<quick_hash<<")."<<endl;
	quick_hash=3;
	cerr << "Setting quickhash to "<<quick_hash<<" instead."<<endl;
      }
      break;
    case 'h':
      if(!sscanf(optarg,"%lu",&hitfilter)){
	cerr << "Could not parse hitfilter specification.\n";
	cerr << program_help();exit(-1);
      }
      break;
    case 'b':
      if(!sscanf(optarg,"%d",&max_hash_bits)){
	cerr << "Could not parse hash size specification.\n";
	cerr << program_help();exit(-1);
      }
      if(max_hash_bits>WORDSIZE || max_hash_bits<1){
	cout << "*** Warning *** "<<max_hash_bits << " hash bits?\n";
	cout << "A hash that big would get you in trouble...\nI'm going to make that "<<(WORDSIZE-3)<<" for you... (and even that may be too big for your system)\n" << endl;
	max_hash_bits=(WORDSIZE-3);
      }
      auto_hashbits=false;
      break;
    case 'H':
      histogram=1;
      if(optarg && sscanf(optarg,"%d",&histogram) && (histogram<0 || histogram>4)){
	cerr << histogram << " is not a valid choice for histogram.\n";
	histogram=2;
	cerr << "Using "<<histogram<<" instead.\n";
      }
      break;
    case 't':
      prefreq=optarg;
      if(prefreq=="ArrayHash")
	hashPref=ArrayHash_t;
      else if(prefreq=="MSetHash" || prefreq=="MultisetHash")
	hashPref=MSetHash_t;
#ifdef MURASAKI_SQLITE3
      else if(prefreq=="SQLiteHash")
	hashPref=SQLiteHash_t;
#endif
#ifdef MURASAKI_MYSQL
      else if(prefreq=="MySQLHash")
	hashPref=MySQLHash_t;
#endif
      else if(prefreq=="EcoHash")
	hashPref=EcoHash_t;
      else cerr << "Unknown hash specification: "<<prefreq<<". Sticking with "<<hashPref_s()<<endl;
      break;
    case 'r':
      parseYesNo(optarg,repeatMask,"repeatmask");break;
    case 's':
      if(!sscanf(optarg,"%d",&rand_seed)){
	cerr << "Could not parse random seed specification (should be a number).\n";
	cerr << program_help();exit(-1);
      }
      user_seed=true;
      break;
    case 'f':
      if(!sscanf(optarg,"%lu",&seedfilter)){
	cerr << "Could not parse seedfilter specification (should be a number).\n";
	cerr << program_help();exit(-1);
      }
      if(seedfilter<0)
	seedfilter=0;
      break;
    case 'm':
      if(!sscanf(optarg,"%lu",&hashfilter)){
	cerr << "Could not parse hashfilter specification (should be a number).\n";
	cerr << program_help();exit(-1);
      }
      if(hashfilter<0)
	hashfilter=0;
      break;
    case 'Y':
      if(!sscanf(optarg,"%lu",&mergeFilter)){
	cerr << "Could not parse mergefilter specification (should be a number).\n";
	cerr << program_help();exit(-1);
      }
      if(mergeFilter<0)
	mergeFilter=0;
      break;
    case 'M':
      double val;
      if(strchr(optarg,'%') && sscanf(optarg,"%lf%%",&val)){
	targetMemory=(int)(sysinfo.totalMemory*val/100);
      }else if(sscanf(optarg,"%lf",&val)){
	targetMemory=(int)(val*1024*1024);
      }else{
	cerr << "Could not parse target memory value (should be a number of gigabytes or a percent).\n";
	cerr << program_help();exit(-1);
      }
      userSetMemory=true;
      break;
    case 'C':parseYesNo(optarg,commutative,"commutative");break;
    case 'i':parseYesNo(optarg,repeatMap,"repeatmap");break;
    case 'B': parseYesNo(optarg,bitscore,"bitscore");break;
    case 'F': parseYesNo(optarg,skipFwd,"skipfwd");break;
    case 'R': parseYesNo(optarg,skipRev,"skiprev");break;
    case '1':case '7': parseYesNo(optarg,skip1to1,"skip1to1");break;
    case 'D': parseYesNo(optarg,doubleHash,"doublehash");break;
    case 'Q': parseYesNo(optarg,hashOnly,"hashonly");break;
    case 'G': parseYesNo(optarg,dumpRegions,"dumpregions");break;
    case 'E': parseYesNo(optarg,dumpEntropy,"dumpentopy");break;
    case 'K': mpi_noCake=true;break;
    case 'L': parseYesNo(optarg,mpi_fileDistro,"mpidistro");break;
    case 'e': useHumanTime=false;break;
    case 'w': mpi_anchorInSpareTime=false;break;
    case 'T': parseYesNo(optarg,retainMembers,"retainMembers");
      if(!retainMembers && tfidf){
	cout << "Warning: tfidf requires --retainmembers=yes. Disabling tfidf."<<endl;
	tfidf=false;
      }
      break;
    case 'g': try {
	anchorProgressCheck=lexical_cast<word>(optarg);
      }catch(bad_lexical_cast& e){
	cerr << "Error: --anchorprogress needs a non-negative integer argument"<<endl;
	exit(-2);
      }
      break;
    case 'U': 
#ifdef MURASAKI_MPI
      if(!mpi_capable)
#endif
	cerr << "Warning: --nobuffer has no effect without MPI."<<endl;
      mpi_maxBuffers=1;break;
    case 'u': 
#ifdef MURASAKI_MPI
      if(!mpi_capable)
#endif
	cerr << "Warning: --buffers has no effect without MPI."<<endl;
#ifdef MURASAKI_MPI
      else
	try {
	  mpi_maxBuffers=lexical_cast<word>(optarg);
	}catch(bad_lexical_cast& e){
	  cerr << "Error: --buffers needs a non-negative integer argument"<<endl;
	}
#endif
      break;
    case 'I': 
#ifdef MURASAKI_MPI
      if(!mpi_capable)
#endif
	cerr << "Warning: --bigfirst has no effect without MPI."<<endl;
      mpi_bigFirst=true;break;
    case 'P': 
#if defined(USE_SHM_MMAP) && defined(MURASAKI_MPI)
      parseYesNo(optarg,use_shm_mmap,"mmap");
      if(use_shm_mmap)
	use_shm_sysv=false; //don't use both
#else
      cerr <<"Warning: The --mmap option requires MMAP support at compile time. (ignoring)"<<endl;
#endif
      break;
    case 'l': 
#ifdef MURASAKI_MPI
      parseYesNo(optarg,mpi_hostbalance,"hostbalance");
#else
      cerr <<"Warning: The --hostbalance option requires MPI support at compile time. (ignoring)"<<endl;
#endif
      break;
    case 'a': 
#ifdef MURASAKI_MPI
      parseYesNo(optarg,mpi_memoryBalance,"memorybalance");
#else
      cerr <<"Warning: The --memorybalance option requires MPI support at compile time. (ignoring)"<<endl;
#endif
      break;
    case 'W': 
#if defined(USE_SHM_MMAP) && defined(MURASAKI_MPI)
      parseYesNo(optarg,mmap_writePerHost,"mmapperhost");
#else
      cerr << "Warning: --mmapperhost requires MMAP support at compile time. (ignoring)"<<endl;
#endif
      break;
    case 'y': 
#ifdef USE_SHM_SYSV
      parseYesNo(optarg,use_shm_sysv,"sysvipc");
      if(use_shm_sysv)
	use_shm_mmap=false; //don't use both
#else
      cerr << "Warning: --sysvipc requires System V IPC support at compile time. (ignoring)"<<endl;
#endif
      break;
    case 'x':
#ifdef MURASAKI_MPI
      if(mpi_capable){
	cerr << "Warning: --extractinorder has no effect when using MPI."<<endl;
	break;
      }
#endif
      parseYesNo(optarg,extractInOrder,"extractinorder");
      break;
    case '<': 
#ifdef MURASAKI_MPI
      parseYesNo(optarg,mpi_distMerge,"distmerge");
      if(mpi_distMerge && !mpi_distCollect){
	cerr << "Warning: --distmerge requires --distcollect (enabling distcollect)"<<endl;
	mpi_distCollect=true;
      }
#else
      cerr <<"Warning: The --distmerge option requires MPI support at compile time. (ignoring)"<<endl;
#endif
      break;
    case '>': 
#ifdef MURASAKI_MPI
      parseYesNo(optarg,mpi_distCollect,"distcollect");
      if(!mpi_distCollect && mpi_distMerge){
	cerr << "Warning: --distmerge requires --distcollect (disabling distmerge)"<<endl;
	mpi_distMerge=false;
      }
#else
      cerr <<"Warning: The --distcollect option requires MPI support at compile time. (ignoring)"<<endl;
#endif
      break;
    case 'k':
      parseYesNo(optarg,tfidf,"tfidf");
      break;
    case 'o':parseYesNo(optarg,reverseOtf,"reverseotf");break;
    case '%':try {
	rifts=lexical_cast<int>(optarg);
      }catch(bad_lexical_cast& e){
	cerr << "Error: --rifts needs a non-negative integer argument"<<endl;
      }
      break;
    default: cout << program_help();exit(1);break;
    }
  }
  if((seedfilter || hashfilter) && !doubleHash){
    cout << "Seedfilter/hashfilter necessitates doublehash: Enabling.\n";
    doubleHash=true;
  }
  if(tfidf && !retainMembers){
    cout << "Warning: tfidf requires --retainmembers=yes, disablign tfidf."<<endl;
    tfidf=false;
  }
}

void parseYesNo(const char* str,bool& toggle,const char *name){
  if(str){
    if(isYes(str))toggle=true;
    else if(isNo(str))toggle=false;
    else {
      cerr << "Please specify either yes/t/true or no/n/false as an argument to --"<<name<<endl;
      cerr << program_help();exit(-1);
    }
  }else{
    toggle=!toggle;
  }
}

bool isYes(const char* str){
  return (!strcasecmp(optarg,"yes") || !strcasecmp(optarg,"y") || !strcasecmp(optarg,"true") || !strcasecmp(optarg,"1"));
}

bool isNo(const char* str){
  return (!strcasecmp(optarg,"no") || !strcasecmp(optarg,"n") || !strcasecmp(optarg,"false") || !strcasecmp(optarg,"0"));
}

void Options::solidify(){ //finalize annoying names and crap
#ifdef MURASAKI_MPI
#ifdef USE_SHM_SYSV
  if(!mpi_capable)
    use_shm_sysv=false;
#endif
#endif
  //transfer options to other modules
  timing_useHumanTime=useHumanTime;

  if((int)rifts>seq_count-2){
    cerr << "Warning: User supplied rifts ("<<rifts<<") is set to high.";
    rifts=seq_count-2;
    cerr << " Setting to "<<rifts<<endl;
  }

  //Any postprocessing
  if(joinNeighbors)
    if(joinSpec<0)
      joinDist=(int)(-1.0*joinSpec*(float)pattern_src.length());
    else
      joinDist=(int)joinSpec;

  //option dependency check.
  if(extractInOrder){
    if(mergeFilter)
      cout << "Warning: MergeFilter has no effect when --extractinorder is enabled."<<endl;
    if(repeatMap)
      cout << "Warning: RepeatMap has no effect when --extractinorder is enabled."<<endl;
    if(rifts)
      cout << "Warning: rifts has no effect when --extractinorder is enabled."<<endl;
  }else
    doubleHash=true;  

  if(hashPref!=MSetHash_t && !doubleHash){
#ifdef MURASAKI_MPI
    if(!mpi_capable)
#endif
      cout << "!Note: Enabling doublehash is highly recommended for "<<hashPref_s()<<endl;
    //    doubleHash=true;
  }

  //make necessary directories
  if(hashCache){
    path cachedir(cache_dir);
    if(exists(cachedir)){
      if(!is_directory(cachedir)){
	cerr << "Cache directory is not a directory...\n";exit(-5);
      }
    }else{
      bool res=false;
      try{res=create_directory(cachedir);}
      catch(exception &e){cerr << "Exceptionated!!!\n"<<e.what()<<endl;exit(-5);}
      if(res)
	cout << "Created cache directory: "
	     <<system_complete(cachedir).string()<<endl;
    }
  }
  
  prefix=output_dir+"/"+alignment_name;
  path dir(output_dir);
  if(dir.empty())
    dir=current_path();
  else {
    if(exists(dir)){
      if(!is_directory(dir)){
	cerr << "Output directory is not a directory...\n";exit(-5);
      }
    }else{
      bool res=false;
      try{res=create_directory(dir);}
      catch(exception &e){cerr << "Exceptionated!!!\n"<<e.what()<<endl;exit(-5);}
      if(res)
	cout << "Created output directory: "<<system_complete(dir).string()<<endl;
    }
  }

  //make up file names
#ifdef MURASAKI_MPI
  if( mpi_id!=0 ){
    char mpi_id_str[81];
    snprintf(mpi_id_str, 80, "%d", mpi_id);
    status_record=prefix+"."+mpi_id_str+".status";
  }
  else
#endif
    status_record=prefix+".status";

  hashStatus_record=prefix+".hashstatus";
  seq_record=prefix+".seqs";
  anchor_record=prefix+".anchors";
  repeat_record=prefix+".repeats";
  pat_record=prefix+".pat";
  region_record=prefix+".regions";
  hashHisto_record=(prefix+".histogram");
  hashDetailed_record=(prefix+".histogram.details");
  options_record=prefix+".options";
  entropy_record=prefix+".entropy";
  anchorProgress_record=prefix+".anchors.progress";

  if(opt.verbose)
    seqreadOptions=SEQO_VERBOSE;
  if(opt.repeatMask)
    seqreadOptions|=SEQO_RMASK;
}

ostream& operator<<(ostream &opf,const Options &a){
  opf << "Verbosity level is: "<<verbose<<endl;
  opf << "Writing to: " << a.prefix<<endl;
  opf << "Pattern is: " << pattern_src<<endl;
  opf << "Max usable hash bits: "<<max_hash_bits<<endl;
  opf << "Hit filter is: "<<a.hitfilter<<endl;
  opf << "Hash type is: "<<a.hashPref_s()<<endl;
  opf << "Double hash: "<<(a.doubleHash ? "on":"off")<<endl;
  opf << "Hashing every: "<<(a.hashSkip)<<" base"<<endl;
  opf << "Histogram detail is: "<<a.histogram<<endl;
  opf << "Hashing function is: "<<a.quick_hash<<endl;
  opf << "Repeats are: "<<(a.repeatMask ? "masked":"unmasked")<<endl;
  if(a.seedfilter)
    opf << "Seedfilter is at: "<<a.seedfilter<<endl;
  if(a.hashfilter)
    opf << "Hashfilter is at: "<<a.hashfilter<<endl;
  if(a.mergeFilter)
    opf << "Mergefilter is at: "<<a.mergeFilter<<endl;
  opf << "Random seed: "<<(a.rand_seed)<<(a.user_seed ? " (custom)":" (automatic)")<<endl;
  if(a.skipFwd || a.skipRev || a.skip1to1)
    opf << "Skip: "<<(a.skipFwd ? "forward ":"")<<(a.skipRev ? "reverse ":"")
	<< (a.skip1to1 ? "1-to-1 ":"") << endl;
  opf << "Bitscore output: "<<(a.bitscore ? "on":"off")<<endl;
  opf << "Leave status records: "<<(a.leaveRecords ? "yes":"no")<<endl;
  opf << "Join neighbors is: "<<(a.joinNeighbors ? (string(" at ")+dstring(a.joinDist)):"off")<<endl;
  opf << "Target memory is: "<<humanMemory(a.targetMemory*1024)<<endl;
  opf << "Retain members is: "<<(a.retainMembers ? "true":"false")<<endl;
  opf << "Commutative is: "<<(a.commutative ? "true":"false")<<endl;
  opf << "Store repeatmap? "<<(a.repeatMap ? "yes":"no")<<endl;
  opf << "Reverse complement data is "<<(a.reverseOtf ? "generated on the fly":"computed at sequence loading")<<endl;
  opf << "Rifts allowed: "<<(a.rifts)<<endl;
#ifdef LARGESEQ_SUPPORT
  bool largeSeqs=true;
#else
  bool largeSeqs=false;
#endif
  opf << "Large sequence support is: "<<(largeSeqs ? "enabled":"disabled");
  opf <<" => seqpos is "<<(sizeof(SeqPos))<<" bytes"<<endl;
#ifdef MURASAKI_MPI  
  opf << "MPI>MPI is: "<<(mpi_capable ? "enabled":"disabled")<<endl;
  if(mpi_capable){
#ifdef USE_SHM_MMAP
  opf << "MPI>MMAP is: "<<(a.use_shm_mmap ? "enabled":"disabled")<<endl;
#endif
#ifdef USE_SHM_SYSV
  opf << "MPI>System V IPC shared memory is: "<<(a.use_shm_sysv ? "enabled":"disabled")<<endl;
#endif
    opf << "MPI>Hashers is: "<<(a.mpi_hashers)<<endl;
    opf << "MPI>np is: "<<mpi_procs<<endl;
    opf << "MPI>MPI-based file distribution is: "<<(a.mpi_fileDistro ? "enabled":"disabled")<<endl;
    opf << "MPI>there is cake? "<<(a.mpi_noCake ? "no":"yes")<<endl;
    opf << "MPI>Wait to anchor? "<<(a.mpi_anchorInSpareTime ? "no":"yes")<<endl;
    opf << "MPI>Max buffers: "<<(a.mpi_maxBuffers ? boost::lexical_cast<string>(a.mpi_maxBuffers):string("unlimited"))<<endl;
    opf << "MPI>Allocate hashers on "<<(a.mpi_bigFirst ? "big":"small")<<" nodes first"<<endl;
    opf << "MPI>Balance hasher allocation? "<<(a.mpi_hostbalance ? "yes":"no")<<endl;
    opf << "MPI>Distributed merging? "<<(a.mpi_distMerge ? "yes":"no")<<endl;
    opf << "MPI>Distributed collection? "<<(a.mpi_distCollect ? "yes":"no")<<endl;
  }
#endif
  opf << program_version();
  return opf;
}

string Options::hashPref_s() const{
  switch(hashPref){
  case ArrayHash_t:return string("Array");
  case MSetHash_t:return string("Multiset");
  case MySQLHash_t:return string("MySQL");
  case SQLiteHash_t: return string("SQLite");
  case EcoHash_t:return string("EcoHash");
  default: return string("Buh????");
  }
  return string("Unknown?????"); //i certainly hope not
}

void Options::loadConfig(string file){
  ifstream cfg(file.c_str());
  //construct a phony argv
  vector<char*> argv;
  argv.push_back("murasaki");//uhhh just dont read that, please
  string tmp;
  char buf[2048];
  while(cfg.good()){
    cfg.getline(buf,sizeof(buf));
    tmp=string(buf);
    while(cfg.fail()){ //for absurdly long inputs
      cfg.getline(buf,sizeof(buf));
      tmp+=buf;
    }
    argv.push_back(new char[tmp.length()]);
    strcpy(argv.back(),tmp.c_str());
  }
  commandline_opts(argv.size(),&argv.front()); //hoho stl rocks
}

string program_help(){
  return string("\
Usage: murasaki [options] seq1 seq2 [seq3 ... ]\n\
Options:\n\
  --directory|-d = output directory (default: output)\n\
  --name|-n      = alignment name (default: test)\n\
  --pattern|-p   = seed pattern (default: 111010010100110111)\n\
                    using the format [<w>:<l>] automatically generates a\n\
                    random pattern of weight <w> and length <l>\n\
  --quickhash|-q = specify a hashing function:\n\
                    1 - don't pack bits to make hash (use first word only)\n\
                    2 - naively use the first hashbits worth of pattern\n\
                    3 - adaptivevely find a good hash (default)\n\
  --hashbits|-b  = use n bit hashes (for n's of 1 to WORDSIZE. default 26)\n\
  --hashtype|-t  = select hash data structure to use:\n\
                    EcoHash    - sub-word packing of hashbits\n\
                    ArrayHash  - balanced malloc/realloc (default)\n\
                    MSetHash   - memory exorbanant, but guaranteeably fast\n\
                    SQLiteHash - use SQLite3\n\
                    MySQLHash  - use MySQL\n\
  --hitfilter|-h = minimum number of hits to be outputted as an anchor\n\
                   (default 1)\n\
  --histogram|-H = histogram computation level: (-H alone implies -H1)\n\
                    0 - no histogram\n\
                    1 - basic bucketsize/bucketcount histogram data (def)\n\
		    2 - bucket-based scores to anchors.detils\n\
                    3 - perbucket count data\n\
		    4 - perbucket + perpattern count data\n\
  --repeatmask|-r= skip repeat masked data (ie: lowercase atgc)\n\
  --seedfilter|-f= skip seeds that occur more than N times\n\
  --hashfilter|-m= like --seedfilter but works on hash keys instead of\n\
                   seeds. May cause some collateral damage to otherwise\n\
                   unique seeds, but it's faster. Also non-sequence-specific\n\
                   so more like at best 1/N the tolerance of seedfilter.\n\
  --rseed|-s     = random number seed for non-deterministic algorithms\n\
                   (ie: the adative hash-finding). If you're doing any\n\
                   performance comparisons, it's probably imperative that you\n\
                   use the same seed for each run of the same settings.\n\
                   Default is obtained from time() (ie: seconds since 1970).\n\
  --skipfwd|-F   = Skip forward facing matches\n\
  --skiprev|-R   = Skip reverse facing matches\n\
  --skip1to1|-1  = Skip matches along the 1:1 line (good for comparing to self)\n\
  --doublehash|-D= Hashes the final sequence twice (once to fill the hash,\n\
                   and once to evaluate matches on it). This is good if your\n\
                   hash happens to be stored externally in a database.\n\
  --hashonly|-Q  = Hash Only. no anchors. just statistics.\n\
  --hashskip|-S  = Hashes every n bases. (Default is 1. ie all)\n\
                   Not supplying any argument increments the skip amount by 1.\n\
  --hashCache|-c = Caches hash tables in the directory. (default: cache/)\n\
  --join|-j      = Join anchors within n bases of eachother (default: 0)\n\
                   Specifying a negative n implies -n*patternLength\n\
  --bitscore|-B  = toggles compututation of a bitscore for all anchors\n\
                   (default is on)\n\
  --memory|-M    = set the target amount of total memory\n\
                    (either in gb or as % total memory)\n\
  --seedterms|-T = toggles retention of seed terms (defaults to on)\n\
                    (these are necessary for computing TF-IDF scores)\n\
  --commutative|-C = forces resuls to be commutative (ie: |a*b|=|b*a|)\n\
                     takes an argument of yes/no (default is yes)\n\
  --sectime|-e   = always display times in seconds\n\
  --repeatmap|-i = toggles keeping of a repeat map when --mergefilter\n\
                   is used (defaults to yes).\n\
  --mergefilter|-Y = filter out matches which would would cause more than N\n\
                     many anchors to be generated from 1 seed (eg. -Y100).\n\
  --extractinorder|-x = toggles whether or not to extract anchors from the\n\
                        hash in sequence order (default is no)\n\
  --tfidf|-k       = perform accurate tfidf scoring from within murasaki\n\
                     (requires extra memory at anchor generation time)\n\
  --reverseotf|-o  = generate reverse complement on the fly (defaults to on)\n\
  --rifts|-%       = allow anchors to skip N sequences (default 0)\n\
")+
#ifdef MURASAKI_MPI
    string("MPI Specific: \n\
  --hashers|-A   = specify the number of processes to be used as hashers\n\
                    (only applies to MPI. If a number between 0 and 1\n\
                     it refers to a % of np)\n\
  --nocake|-K    = perform hashing locally on each storage node rather than\n\
                   sending it over the network (helpful for slow networks)\n\
  --mpidistro|-L  = toggles use of MPI to distribute sequence data over\n\
                    (if the sequence is available on local disk on each\n\
                     node then turning this off may increase performance)\n\
  --waittoanchor|-w = postpone actual anchor computation until all location\n\
                      sets have been received.\n\
  --buffers|-u  = maximum number of unfinished buffers to allow while\n\
                     message passing (0 means unlimited)\n\
  --nobuffers|-U   = same as --buffers=1\n\
  --bigfirst|-I    = assign hashers to large memory nodes first\n\
  --hostbalance|-l = if yes (default): spread out hashers evenly among hosts\n\
                     if no: ignore host name when assigning jobs\n\
  --memorybalance|-a = if yes (deafult): balance hash storage between nodes\n\
                       based on the amount of available ram.\n\
                       if no: distribute storage evently.\n\
  --distmerge|-< = if yes (default): during the merge step, send seeds to\n\
                   all participating hashers.\n\
                   if no: send all seeds to one node only\n\
  --distcollect|-> = if yes (default): collect anchor data from all hashers\n\
                   if no: send all seeds to the final assembly node only\n\
")+
#endif
#ifdef USE_SHM_MMAP
    string("\
  --mmap|-P      = use filebacked mmap() to store sequence data\n\
                    (saves memory when one host runs multiple nodes)\n\
  --mmapperhost|-W = create files for mmap() on each host\n\
                     (ie: for when sequence sources aren't stored on NFS)\n\
")+
#endif
#ifdef USE_SHM_SYSV
    string("\
  --sysvipc|-V   = use System V IPC to negotiate shared memory regions\n\
                    (saves memory when one host runs multiple nodes)\n\
")+
#endif
    string("\n\
Debugging options:\n\
  --leaverecords|-J= don't erase .status and -mpiout files on successful\n\
                     completion (default is to erase them)\n\
  --dumpentropy|-E = dump entropy data for each anchor\n\
  --dumpregions|-G = dump list of comparable regions for each sequence\n\
                     (ie: the areas between N's)\n\
  --anchorprogress|-g = log the number of anchors in memory and the amount\n\
                        of free memory every N location-sets processed\n\
\n\
...and of course\n\
  --verbose|-v   = increases verbosity\n\
  --version|-V   = prints version information and quits\n\
  --help|-?      = prints this help message and quits\n\
\n")+platformInfo()+program_version();
}
